#!/usr/bin/python3
import ctypes
from multiprocessing.sharedctypes import Value
from select import select
from ctypes import *
from enum import Enum
import socket
import psutil
import json
import threading
import re
import time
from pydbus import SystemBus
import dbus
import syslog
import os
import schedule
import gi
gi.require_version('Gtk', '3.0')
from gi.repository import GLib, Gtk, Gio

BUS_NAME = "com.kylin.processMonitor"
OBJ_NAME = "/com/kylin/processMonitor"
IFACE_NAME_MONITOR = "com.kylin.processMonitor.startend"


fun_ppro = 5
class kysec_message_type(Enum):
    PPRO_ADD = 9
    PPRO_DEL = 10
    PPRO_PID = 12

class kysec_ppro_type(Enum):
    PPRO_SUBS = 1
    PPRO_KEY = 2
    PPRO_MONITER = 3

class kysec_func_status(Enum):
	FUNC_DISABLED = 0
	FUNC_ENABLED = 1

class kysec_ppro_kernel(Structure):
    _fields_ = [
        ('path', ctypes.c_char * 4096),
        ('hash', ctypes.c_char * 65),
        ('type', ctypes.c_int)
    ]

class KeyProcessMonitor:
    def __init__(self):
        self.extend_so = CDLL("libkysec_extend.so.0")

    def get_pids_by_path(self, path):
        pids = []
        for proc in psutil.process_iter(['pid', 'exe']):
            try:
                if proc.info['exe'] and os.path.samefile(proc.info['exe'], path):
                    pids.append(proc.info['pid'])
            except (psutil.NoSuchProcess, FileNotFoundError, PermissionError):
                continue
        return pids

    def kysec_process_enable(self):
        ret = self.extend_so.kysec_getstatus()
        if ret != 2:
            syslog.syslog(syslog.LOG_DEBUG, "kysec get status error")
            return -1

        ret = self.extend_so.kysec_get_func_status(c_int(fun_ppro))
        if ret == 1 or ret == 2:
            return 0

        if ret == 0:
            ret = self.extend_so.kysec_set_func_status(c_int(fun_ppro), c_int(kysec_func_status.FUNC_ENABLED.value))
            if ret != 0:
                syslog.syslog(syslog.LOG_DEBUG, "kysec set func status error")
                return -1

            ret = self.extend_so.kysec_conf_set("kysec_ppro", c_int(kysec_func_status.FUNC_ENABLED.value))
            if ret != 0:
                syslog.syslog(syslog.LOG_DEBUG, "kysec set conf error")
                return -1

        return 0

    def kysec_ppro_add(self, path):
        if not path:
            syslog.syslog(syslog.LOG_DEBUG, "path is invalid")
            return -1

        self.extend_so.kysec_ppro_read_node_with_path.restype = POINTER(kysec_ppro_kernel)
        ppro = self.extend_so.kysec_ppro_read_node_with_path(c_char_p(bytes(path, encoding = "utf8")))
        if ppro:
            ret = self.extend_so.kysec_ppro_update(c_char_p(bytes(path, encoding = "utf8")), c_int(kysec_ppro_type.PPRO_MONITER.value))
        else:
            ret = self.extend_so.kysec_ppro_add(c_char_p(bytes(path, encoding = "utf8")), c_int(kysec_ppro_type.PPRO_MONITER.value))
        if ret != 0:
            syslog.syslog(syslog.LOG_DEBUG, "kysec ppro add error, path = {}, ret = {}".format(path, ret))
            return -1

        self.extend_so.kysec_ppro_get_path_hash.restype = c_char_p
        hash = self.extend_so.kysec_ppro_get_path_hash(c_char_p(bytes(path, encoding = "utf8")))
        if not hash:
            syslog.syslog(syslog.LOG_DEBUG, "kysec ppro get hash error, path = {}".format(path))
            return -1

        ret = self.extend_so.kysec_ppro_communicate_with_kernel(c_int(kysec_message_type.PPRO_ADD.value), hash, kysec_ppro_type.PPRO_MONITER.value)
        syslog.syslog(syslog.LOG_DEBUG, "add type = {}, path = {}, hash = {}, ret = {}".format(kysec_ppro_type.PPRO_MONITER.value, path, str(hash, "utf8"), ret))
        if ret != 0:
            syslog.syslog(syslog.LOG_DEBUG, "kysec add ppro communicate with kernel error, path = {}, ret = {}".format(path, ret))
            return -1

        pids = self.get_pids_by_path(path)
        for pid in pids:
            if pid > 0:
                ret = self.extend_so.kysec_ppro_communicate_with_kernel(c_int(kysec_message_type.PPRO_PID.value), c_char_p(bytes("add", encoding = "utf8")), c_int(pid))
                if ret != 0:
                    syslog.syslog(syslog.LOG_DEBUG, "send ppro data to kernel failed, path = {}, pid = {}, ret = {}".format(path, pid, ret))
                    return -1

        return 0

    def kysec_ppro_delete(self, path):
        if not path:
            syslog.syslog(syslog.LOG_DEBUG, "path is invalid")
            return -1

        self.extend_so.kysec_ppro_read_node_with_path.restype = POINTER(kysec_ppro_kernel)
        ppro = self.extend_so.kysec_ppro_read_node_with_path(c_char_p(bytes(path, encoding = "utf8")))
        if not ppro:
            return 0

        ret = self.extend_so.kysec_ppro_remove(c_char_p(bytes(path, encoding = "utf8")))
        if ret != 0:
            syslog.syslog(syslog.LOG_DEBUG, "kysec ppro remove error, path = {}, ret = {}".format(path, ret))
            return -1

        ret = self.extend_so.kysec_ppro_communicate_with_kernel(c_int(kysec_message_type.PPRO_DEL.value), ppro.contents.hash, ppro.contents.type)
        syslog.syslog(syslog.LOG_DEBUG, "del type = {}, path = {}, hash = {}, ret = {}".format(ppro.contents.type, str(ppro.contents.path, "utf8"), str(ppro.contents.hash, "utf8"), ret))
        if ret != 0:
            syslog.syslog(syslog.LOG_DEBUG, "kysec remove ppro communicate with kernel error, path = {}, ret = {}".format(path, ret))
            return -1

        pids = self.get_pids_by_path(path)
        for pid in pids:
            if pid > 0:
                ret = self.extend_so.kysec_ppro_communicate_with_kernel(c_int(kysec_message_type.PPRO_PID.value), c_char_p(bytes("del", encoding = "utf8")), c_int(pid))
                if ret != 0:
                    syslog.syslog(syslog.LOG_DEBUG, "send ppro data to kernel failed, path = {}, pid = {}, ret = {}".format(path, pid, ret))
                    return -1
        return 0

    def kysec_ppro_get(self):
        self.extend_so.kysec_ppro_read_all_data.restype = POINTER(kysec_ppro_kernel)
        count = c_int(0)
        list = self.extend_so.kysec_ppro_read_all_data(byref(count))
        return list, count.value

    def kysec_ppro_clear(self):
        list, count = self.kysec_ppro_get()
        for i in range(0, count):
            # print(list[i].type, list[i].path, list[i].hash)
            if list[i].type == kysec_ppro_type.PPRO_MONITER.value:
                ret = self.kysec_ppro_delete(str(list[i].path, "utf8"))
                if ret != 0:
                    syslog.syslog(syslog.LOG_DEBUG, "kysec ppro delete error")
                    return -1
        return 0

class processInfo:
    def __init__(self, pid):
        self.pid = pid
        self.status = "run"
        self.command = ""
        self.apath = ""

        try:
            absolutepath = "/proc/" + self.pid + "/exe"
            self.apath = os.readlink(absolutepath)
            self.command = self.apath[self.apath.rfind("/")+1:]
        except Exception:
            syslog.syslog(syslog.LOG_DEBUG, "application procfs not exist.")
            self.command = ""
            self.apath = ""

class ParentService(dbus.service.Object):
    def __init__(self):
        bus = dbus.SystemBus()
        bus_name = dbus.service.BusName(BUS_NAME, bus=bus)
        dbus.service.Object.__init__(self, bus_name, OBJ_NAME)
class OfficeProcessMonitor(ParentService):
    """
    <node>
        <interface name='com.kylin.processMonitor.startend'>
            <method name='setProcess' >
                <arg type='s' direction='in' name='list'/>
                <arg type='i' direction='out' name='ret'/>
            </method>
            <method name='getProcess' >
                <arg type='as' direction='out' name='list'/>
            </method>

            <method name='getCommandbyPid' >
                <arg type='s' direction='in' name='pid' />
                <arg type='s' direction='out' name='command' />
            </method>
        </interface>
    </node>
    """

    def __init__(self):
        try:
            ParentService.__init__(self)
            self.keyProcessListPath = "/etc/key-process-list"
            self.auditlogPath = "/var/log/audit/audit.log"
            self.processList = []  #保存关键进程信息，通过pid获取应用名称信息
            self.officeExecList = self.getProcess() #保存配置的关键应用

            #日志监控
            logMonitorThread = threading.Thread(target=self.logMonitor, args=(self,))
            logMonitorThread.start()
            print("start processMonitor...")

            #定时任务
            schedule.every().day.at('02:00').do(self.dailySend)
        except Exception as e:
            syslog.syslog(syslog.LOG_ERR, f"processMonitor init error: {e}")
            exit(-1)

    @property
    def get_host_name(self):
        """获取当前用户名"""
        try:
            bus2 = dbus.SystemBus()
            proxy = bus2.get_object("com.kylin.kydevmonit.hedronclient","/kydevmonit/hedronclient")
            username = proxy.get_current_user()
            if(username == '' or username == ' '):
                username = '未登录'
        except:
            syslog.syslog(syslog.LOG_DEBUG, "call hedronclient get_current_user error")
            username = '未登录'
        return  username

    @property
    def get_host_time(self):
        """获取本机时间
        return: cur_time
        """
        cur_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        return cur_time

    @property
    def get_host_ip(self):
        """
        查询本机ip地址
        :return: ip
        """
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect(('8.8.8.8', 80))
            ip = s.getsockname()[0]
        except Exception:
            return "unknown"
        return ip

    def log_transmit(self, name, message: str):
        time = self.get_host_time
        hostname = self.get_host_name
        ip = self.get_host_ip
        log = {
            "time":time,
            "hostname":hostname,
            "ip":ip,
            "name":name,
            "lv":"info",
            "message":message
        }
        log_message = json.dumps(log, ensure_ascii=False)
        try:
            self.log = SystemBus().get("org.log.sys_transmit", "/org/log/sys_transmit")
            ret = self.log.log_transmit(log_message)
            syslog.syslog(syslog.LOG_DEBUG, "sent log message = {}, ret = {}".format(log_message, ret))
        except Exception as e:
            syslog.syslog(syslog.LOG_ERR, f"send log message failed: {e}")

    def saveProcessList(self, list):
        f = open(self.keyProcessListPath, "w")
        for path in self.officeExecList:
            f.writelines(path + "\n")
            syslog.syslog(syslog.LOG_DEBUG, "save key process path = {}".format(path))
        f.close()

    @dbus.service.method(dbus_interface=IFACE_NAME_MONITOR, in_signature="s", out_signature="s", sender_keyword='sender')
    def getCommandbyPid(self, pid, sender=None) -> str:
        for pi in self.processList:
            if(pi.pid == pid):
                return pi.command
        return ""

    @dbus.service.method(dbus_interface=IFACE_NAME_MONITOR, in_signature="", out_signature="as", sender_keyword='sender')
    def getProcess(self, sender=None):
        path = []
        try:
            f = open(self.keyProcessListPath, "r")
            path = f.read().splitlines()
            syslog.syslog(syslog.LOG_DEBUG, "get key process path = {}".format(path))
            f.close()
        except Exception:
            syslog.syslog(syslog.LOG_DEBUG, "{} does not exist".format(self.keyProcessListPath))
        return path

    @dbus.service.method(dbus_interface=IFACE_NAME_MONITOR, in_signature="s", out_signature="i", sender_keyword='sender')
    def setProcess(self, list, sender=None):
        tmplist =  self.officeExecList
        try:
            self.officeExecList = json.loads(list)
            syslog.syslog(syslog.LOG_DEBUG, "set Process list = {}".format(self.officeExecList))
        except Exception:
            syslog.syslog(syslog.LOG_DEBUG, "set process list is invalid")
            return -1

        controller = KeyProcessMonitor()
        ret = controller.kysec_process_enable()
        if ret != 0:
            syslog.syslog(syslog.LOG_DEBUG, "enable kysec process protect failed")
            return ret

        for path in tmplist:
            ret = controller.kysec_ppro_delete(path)
            if ret != 0:
                syslog.syslog(syslog.LOG_DEBUG, "delete kysec process {} failed.".format(path))
        
        for path in self.officeExecList:
            ret = controller.kysec_ppro_add(path)
            if ret != 0:
                syslog.syslog(syslog.LOG_DEBUG, "add {} failed".format(path))
        self.saveProcessList(self.officeExecList)
        return ret

    @staticmethod
    def logMonitor(self):
        syslog.syslog(syslog.LOG_DEBUG, "logMonitor start")
        cur_file_size = 0
        new_file_size = 0
        file_handle = None

        while True:
            # 零点轮询
            schedule.run_pending()

            # 文件不存在
            if not os.path.exists(self.auditlogPath):
                syslog.syslog(syslog.LOG_DEBUG, "Audit log file does not exist. ")
                time.sleep(10)
                continue

            # 文件监控
            new_file_size = os.path.getsize(self.auditlogPath)
            if file_handle is None:
                try:
                    file_handle = open(self.auditlogPath, 'r', encoding="utf-8")
                    file_handle.seek(0, 2)
                    cur_file_size = new_file_size
                except IOError as e:
                    syslog.syslog(syslog.LOG_ERR, f"Error opening file: {e}")
                    time.sleep(1)
                    continue

            if new_file_size < cur_file_size:
                syslog.syslog(syslog.LOG_DEBUG, "reload audit.log")
                file_handle.close()
                file_handle = open(self.auditlogPath, 'r', encoding="utf-8")
                file_handle.seek(0, 2)
                cur_file_size = new_file_size

            try:
                data = file_handle.readlines()
                for line in data:
                    if line == '\n':    #空行不处理
                        continue
                    self.parseLog(line)
            except Exception as e:
                syslog.syslog(syslog.LOG_ERR, f"read log error: {e}")
                continue

            cur_file_size = new_file_size
            time.sleep(1)

    def sendOfficeInfo(self, pi:processInfo):
        message = "status:" + pi.status + ", comm:" + pi.apath
        self.log_transmit("processmonitor", message)

    def parseLog(self, line):
        try:
            if "{ run }" in line:
                syslog.syslog(syslog.LOG_DEBUG, "get run log: {}".format(line))
                p = 'type=.* msg=audit\(([0-9]{10}).*\).*?{ run } for msg="pid=(.*) comm=.*" loginuid=(.*) session=.*'
                info = re.findall(p, line)

                pi = processInfo(info[0][1])
                if pi.apath in self.officeExecList:
                    self.sendOfficeInfo(pi)

                if pi.apath != "":
                    self.processList.append(pi)

            if "{ exit }" in line:
                syslog.syslog(syslog.LOG_DEBUG, "get exit log: {}".format(line))
                p = 'type=.* msg=audit\(([0-9]{10}).*\).*?{ exit } for msg="pid=(.*) comm=.*'
                info = re.findall(p, line)

                for pi in self.processList:
                    if pi.pid != info[0][1]:
                        continue

                    pi.status = "exit"
                    if pi.apath in self.officeExecList:
                        self.sendOfficeInfo(pi)
                        break
        except Exception as e:
            syslog.syslog(syslog.LOG_ERR, f"analyse log error: {e}")

    def dailySend(self):
        #零点发送还在运行的关键应用、重置processList
        syslog.syslog(syslog.LOG_DEBUG, "dailySend: get officeExecList = {}".format(self.officeExecList))
        tmplist = []
        for pi in self.processList:
                if os.path.exists("/proc/" + pi.pid):
                    tmplist.append(pi)
                    if(pi.apath in self.officeExecList):
                        syslog.syslog(syslog.LOG_DEBUG, "dailySend: get run pid = {}, status = {}, comm = {}, apath = {}".format(pi.pid, pi.status, pi.command, pi.apath))
                        self.sendOfficeInfo(pi)
        self.processList = tmplist
